@@ -34,13 +34,13 @@
 static char doc[] = "OCI runtime";
 
 enum
-  {
-    OPTION_CONSOLE_SOCKET = 1000,
-    OPTION_PID_FILE,
-    OPTION_NO_SUBREAPER,
-    OPTION_NO_NEW_KEYRING,
-    OPTION_PRESERVE_FDS
-  };
+{
+  OPTION_CONSOLE_SOCKET = 1000,
+  OPTION_PID_FILE,
+  OPTION_NO_SUBREAPER,
+  OPTION_NO_NEW_KEYRING,
+  OPTION_PRESERVE_FDS
+};
 
 struct kill_options_s
 {
@@ -50,12 +50,12 @@ struct kill_options_s
 
 static struct kill_options_s kill_options;
 
-static struct argp_option options[] =
-  {
-   {"all", 'a', 0, 0, "kill all the processes", 0 },
-   {"regex", 'r', 0, 0, "the specified CONTAINER is a regular expression (kill multiple containers)", 0 },
-   { 0, }
-  };
+static struct argp_option options[]
+    = { { "all", 'a', 0, 0, "kill all the processes", 0 },
+        { "regex", 'r', 0, 0, "the specified CONTAINER is a regular expression (kill multiple containers)", 0 },
+        {
+            0,
+        } };
 
 static char args_doc[] = "kill CONTAINER [SIGNAL]";
 
@@ -89,7 +89,9 @@ crun_command_kill (struct crun_global_arguments *global_args, int argc, char **a
 {
   int first_arg, signal, ret;
 
-  libcrun_context_t crun_context = {0, };
+  libcrun_context_t crun_context = {
+    0,
+  };
 
   argp_parse (&run_argp, argc, argv, ARGP_IN_ORDER, &first_arg, &kill_options);
   crun_assert_n_args (argc - first_arg, 1, 2);
@@ -101,7 +103,7 @@ crun_command_kill (struct crun_global_arguments *global_args, int argc, char **a
   signal = SIGTERM;
   if (argc - first_arg > 1)
     {
-      signal = str2sig (argv[first_arg + 1]);
+      signal = libcrun_str2sig (argv[first_arg + 1]);
       if (UNLIKELY (signal < 0))
         libcrun_fail_with_error (0, "unknown signal %s", argv[first_arg + 1]);
     }
@@ -124,10 +126,9 @@ crun_command_kill (struct crun_global_arguments *global_args, int argc, char **a
           {
             ret = libcrun_container_kill (&crun_context, it->name, signal, err);
             if (UNLIKELY (ret < 0))
-              crun_error_write_warning_and_release (stderr, &err);
+              libcrun_error_write_warning_and_release (stderr, &err);
           }
 
-
       libcrun_free_containers_list (list);
       regfree (&re);
       return 0;
