@@ -76,16 +76,16 @@ detect_cgroup_mode (libcrun_error_t *err)
   struct statfs stat;
   int ret;
 
-  ret = statfs ("/sys/fs/cgroup", &stat);
+  ret = statfs (CGROUP_ROOT, &stat);
   if (ret < 0)
-    return crun_make_error (err, errno, "statfs '/sys/fs/cgroup'");
+    return crun_make_error (err, errno, "statfs '" CGROUP_ROOT "'");
   if (stat.f_type == CGROUP2_SUPER_MAGIC)
     return CGROUP_MODE_UNIFIED;
   if (stat.f_type != TMPFS_MAGIC)
-    return crun_make_error (err, 0, "invalid file system type on '/sys/fs/cgroup'");
-  ret = statfs ("/sys/fs/cgroup/unified", &stat);
+    return crun_make_error (err, 0, "invalid file system type on '" CGROUP_ROOT "'");
+  ret = statfs (CGROUP_ROOT "/unified", &stat);
   if (ret < 0 && errno != ENOENT)
-    return crun_make_error (err, errno, "statfs '/sys/fs/cgroup/unified'");
+    return crun_make_error (err, errno, "statfs '" CGROUP_ROOT "/unified'");
   if (ret < 0)
     return CGROUP_MODE_LEGACY;
   return stat.f_type == CGROUP2_SUPER_MAGIC ? CGROUP_MODE_HYBRID : CGROUP_MODE_LEGACY;
@@ -198,9 +198,12 @@ write_controller_file (const char *path, int controllers_to_enable, libcrun_erro
   int ret;
 
   controllers_len = xasprintf (
-      &controllers, "%s %s %s %s %s %s", (controllers_to_enable & CGROUP_CPU) ? "+cpu" : "",
-      (controllers_to_enable & CGROUP_IO) ? "+io" : "", (controllers_to_enable & CGROUP_MEMORY) ? "+memory" : "",
-      (controllers_to_enable & CGROUP_PIDS) ? "+pids" : "", (controllers_to_enable & CGROUP_CPUSET) ? "+cpuset" : "",
+      &controllers, "%s %s %s %s %s %s",
+      (controllers_to_enable & CGROUP_CPU) ? "+cpu" : "",
+      (controllers_to_enable & CGROUP_IO) ? "+io" : "",
+      (controllers_to_enable & CGROUP_MEMORY) ? "+memory" : "",
+      (controllers_to_enable & CGROUP_PIDS) ? "+pids" : "",
+      (controllers_to_enable & CGROUP_CPUSET) ? "+cpuset" : "",
       (controllers_to_enable & CGROUP_HUGETLB) ? "+hugetlb" : "");
 
   xasprintf (&subtree_control, "%s/cgroup.subtree_control", path);
@@ -259,14 +262,14 @@ enable_controllers (const char *path, libcrun_error_t *err)
 
   xasprintf (&tmp_path, "%s/", path);
 
-  ret = read_available_controllers ("/sys/fs/cgroup", err);
+  ret = read_available_controllers (CGROUP_ROOT, err);
   if (UNLIKELY (ret < 0))
     return ret;
 
   controllers_to_enable = ret;
 
   /* Enable all possible controllers in the root cgroup.  */
-  ret = write_controller_file ("/sys/fs/cgroup", controllers_to_enable, err);
+  ret = write_controller_file (CGROUP_ROOT, controllers_to_enable, err);
   if (UNLIKELY (ret < 0))
     return ret;
 
@@ -277,7 +280,10 @@ enable_controllers (const char *path, libcrun_error_t *err)
 
       *it = '\0';
 
-      xasprintf (&cgroup_path, "/sys/fs/cgroup%s", tmp_path);
+      ret = append_paths (&cgroup_path, err, CGROUP_ROOT, tmp_path, NULL);
+      if (UNLIKELY (ret < 0))
+        return ret;
+
       ret = mkdir (cgroup_path, 0755);
       if (UNLIKELY (ret < 0 && errno != EEXIST))
         return crun_make_error (err, errno, "create `%s`", cgroup_path);
@@ -373,6 +379,75 @@ initialize_cpuset_subsystem_rec (char *path, size_t path_len, char *cpus, char *
   return 0;
 }
 
+static int
+check_cgroup_v2_controller_available_wrapper (int ret, int cgroup_dirfd, const char *name, libcrun_error_t *err)
+{
+  if (ret == 0 || err == NULL)
+    return 0;
+
+  errno = crun_error_get_errno (err);
+
+  /* If the file is not found, try to give a more meaningful error message.  */
+  if (errno == ENOENT || errno == EPERM || errno == EACCES)
+    {
+      cleanup_free char *controllers = NULL;
+      libcrun_error_t tmp_err = NULL;
+      cleanup_free char *key = NULL;
+      char *saveptr = NULL;
+      bool found = false;
+      const char *token;
+      char *it;
+
+      /* Check if the specified controller is enabled.  */
+      key = xstrdup (name);
+
+      it = strchr (key, '.');
+      if (it == NULL)
+        {
+          crun_error_release (err);
+          return crun_make_error (err, 0, "the specified key has not the form CONTROLLER.VALUE `%s`",
+                                  name);
+        }
+      *it = '\0';
+
+      /* cgroup. files are not part of a controller.  Return the original error.  */
+      if (strcmp (key, "cgroup") == 0)
+        return ret;
+
+      /* If the cgroup.controllers file cannot be read, return the original error.  */
+      if (read_all_file_at (cgroup_dirfd, "cgroup.controllers", &controllers, NULL, &tmp_err) < 0)
+        {
+          crun_error_release (&tmp_err);
+          return ret;
+        }
+      for (token = strtok_r (controllers, " \n", &saveptr); token; token = strtok_r (NULL, " \n", &saveptr))
+        {
+          if (strcmp (token, key) == 0)
+            {
+              found = true;
+              break;
+            }
+        }
+      if (! found)
+        {
+          crun_error_release (err);
+          return crun_make_error (err, 0, "the requested cgroup controller `%s` is not available", key);
+        }
+    }
+  return ret;
+}
+
+static int
+write_file_and_check_controllers_at (bool cgroup2, int dirfd, const char *name, const void *data, size_t len, libcrun_error_t *err)
+{
+  int ret;
+
+  ret = write_file_at (dirfd, name, data, len, err);
+  if (cgroup2)
+    return check_cgroup_v2_controller_available_wrapper (ret, dirfd, name, err);
+  return ret;
+}
+
 static int
 initialize_cpuset_subsystem (const char *path, libcrun_error_t *err)
 {
@@ -416,11 +491,14 @@ move_process_to_cgroup (pid_t pid, const char *subsystem, const char *path, libc
 {
   cleanup_free char *cgroup_path_procs = NULL;
   char pid_str[16];
+  int ret;
+
+  ret = append_paths (&cgroup_path_procs, err, CGROUP_ROOT, subsystem ? subsystem : "", path ? path : "", "cgroup.procs", NULL);
+  if (UNLIKELY (ret < 0))
+    return ret;
 
   sprintf (pid_str, "%d", pid);
 
-  xasprintf (&cgroup_path_procs, "/sys/fs/cgroup/%s%s/cgroup.procs", subsystem ? subsystem : "", path ? path : "");
-
   return write_file (cgroup_path_procs, pid_str, strlen (pid_str), err);
 }
 
@@ -431,7 +509,10 @@ enter_cgroup_subsystem (pid_t pid, const char *subsystem, const char *path, bool
   cleanup_free char *cgroup_path = NULL;
   int ret;
 
-  xasprintf (&cgroup_path, "/sys/fs/cgroup/%s%s", subsystem, path ? path : "");
+  ret = append_paths (&cgroup_path, err, CGROUP_ROOT, subsystem ? subsystem : "", path ? path : "", NULL);
+  if (UNLIKELY (ret < 0))
+    return ret;
+
   if (create_if_missing)
     {
       ret = crun_ensure_directory (cgroup_path, 0755, false, err);
@@ -513,30 +594,40 @@ static int
 chown_cgroups (const char *path, uid_t uid, gid_t gid, libcrun_error_t *err)
 {
   cleanup_free char *cgroup_path = NULL;
-  cleanup_dir DIR *dir = NULL;
-  struct dirent *next;
+  cleanup_free char *delegate = NULL;
+  cleanup_close int dfd = -1;
+  size_t delegate_size;
+  char *saveptr = NULL;
+  char *name;
   int ret;
-  int dfd;
 
-  xasprintf (&cgroup_path, "/sys/fs/cgroup/%s", path);
+  ret = append_paths (&cgroup_path, err, CGROUP_ROOT, path, NULL);
+  if (UNLIKELY (ret < 0))
+    return ret;
 
-  dir = opendir (cgroup_path);
-  if (UNLIKELY (dir == NULL))
-    return crun_make_error (err, errno, "cannot opendir `%s`", cgroup_path);
+  dfd = open (cgroup_path, O_PATH);
 
-  dfd = dirfd (dir);
-
-  for (next = readdir (dir); next; next = readdir (dir))
+  ret = read_all_file ("/sys/kernel/cgroup/delegate", &delegate, &delegate_size, err);
+  if (UNLIKELY (ret < 0))
     {
-      const char *name = next->d_name;
-
-      /* do not chown the parent directory, but chown the current one.  */
-      if (name[0] == '.' && name[1] == '.' && name[2] == '\0')
-        continue;
+      if (crun_error_get_errno (err) == ENOENT)
+        {
+          crun_error_release (err);
+          return 0;
+        }
+      return ret;
+    }
 
+  for (name = strtok_r (delegate, "\n", &saveptr); name; name = strtok_r (NULL, "\n", &saveptr))
+    {
       ret = fchownat (dfd, name, uid, gid, AT_SYMLINK_NOFOLLOW);
       if (UNLIKELY (ret < 0))
-        return crun_make_error (err, errno, "cannot chown `%s/%s`", cgroup_path, name);
+        {
+          if (errno == ENOENT)
+            continue;
+
+          return crun_make_error (err, errno, "cannot chown `%s/%s`", cgroup_path, name);
+        }
     }
 
   return 0;
@@ -615,7 +706,7 @@ enter_cgroup_v1 (int cgroup_mode, pid_t pid, const char *path, bool create_if_mi
       if (cgroup_mode == CGROUP_MODE_LEGACY && strcmp (subsystems[i], "unified") == 0)
         continue;
 
-      sprintf (subsystem_path, "/sys/fs/cgroup/%s", subsystems[i]);
+      sprintf (subsystem_path, CGROUP_ROOT "/%s", subsystems[i]);
       ret = crun_path_exists (subsystem_path, err);
       if (UNLIKELY (ret < 0))
         return ret;
@@ -650,7 +741,10 @@ enter_cgroup_v2 (pid_t pid, pid_t init_pid, const char *path, bool create_if_mis
 
   sprintf (pid_str, "%d", pid);
 
-  xasprintf (&cgroup_path, "/sys/fs/cgroup/%s", path);
+  ret = append_paths (&cgroup_path, err, CGROUP_ROOT, path, NULL);
+  if (UNLIKELY (ret < 0))
+    return ret;
+
   if (create_if_missing)
     {
       ret = crun_ensure_directory (cgroup_path, 0755, false, err);
@@ -658,7 +752,10 @@ enter_cgroup_v2 (pid_t pid, pid_t init_pid, const char *path, bool create_if_mis
         return ret;
     }
 
-  xasprintf (&cgroup_path_procs, "/sys/fs/cgroup/%s/cgroup.procs", path);
+  ret = append_paths (&cgroup_path_procs, err, cgroup_path, "cgroup.procs", NULL);
+  if (UNLIKELY (ret < 0))
+    return ret;
+
   ret = write_file (cgroup_path_procs, pid_str, strlen (pid_str), err);
   if (LIKELY (ret >= 0))
     return ret;
@@ -697,7 +794,9 @@ enter_cgroup_v2 (pid_t pid, pid_t init_pid, const char *path, bool create_if_mis
       if (cgroup_crun_exec_path == NULL)
         xasprintf (&cgroup_crun_exec_path, "%s/crun-exec", path);
 
-      xasprintf (&cgroup_sub_path_procs, "/sys/fs/cgroup/%s/cgroup.procs", cgroup_crun_exec_path);
+      ret = append_paths (&cgroup_sub_path_procs, err, CGROUP_ROOT, cgroup_crun_exec_path, "cgroup.procs", NULL);
+      if (UNLIKELY (ret < 0))
+        return ret;
 
       ret = write_file (cgroup_sub_path_procs, pid_str, strlen (pid_str), err);
       if (UNLIKELY (ret < 0))
@@ -818,10 +917,14 @@ systemd_finalize (struct libcrun_cgroup_args *args, const char *suffix, libcrun_
       if (UNLIKELY (to == NULL))
         return crun_make_error (err, -1, "cannot parse /proc/self/cgroup");
       *to = '\0';
-      if (suffix)
-        xasprintf (path, "%s/%s", from, suffix);
-      else
+      if (suffix == NULL)
         *path = xstrdup (from);
+      else
+        {
+          ret = append_paths (path, err, from, suffix, NULL);
+          if (UNLIKELY (ret < 0))
+            return ret;
+        }
       *to = '\n';
     }
   else
@@ -835,20 +938,28 @@ systemd_finalize (struct libcrun_cgroup_args *args, const char *suffix, libcrun_
       if (UNLIKELY (to == NULL))
         return crun_make_error (err, -1, "cannot parse /proc/self/cgroup");
       *to = '\0';
-      if (suffix)
-        xasprintf (path, "%s/%s", from, suffix);
-      else
+      if (suffix == NULL)
         *path = xstrdup (from);
+      else
+        {
+          ret = append_paths (path, err, from, suffix, NULL);
+          if (UNLIKELY (ret < 0))
+            return ret;
+        }
       *to = '\n';
     }
 
   if (cgroup_mode == CGROUP_MODE_UNIFIED)
     {
+      bool must_chown = false;
+
       if (suffix)
         {
           cleanup_free char *dir = NULL;
 
-          xasprintf (&dir, "/sys/fs/cgroup%s", *path);
+          ret = append_paths (&dir, err, CGROUP_ROOT, *path, NULL);
+          if (UNLIKELY (ret < 0))
+            return ret;
 
           /* On cgroup v2, processes can be only in leaf nodes.  If a suffix is used,
              move the process immediately to the new location before enabling
@@ -858,18 +969,23 @@ systemd_finalize (struct libcrun_cgroup_args *args, const char *suffix, libcrun_
           if (UNLIKELY (ret < 0))
             return ret;
 
-          ret = chown_cgroups (*path, args->root_uid, args->root_gid, err);
-          if (UNLIKELY (ret < 0))
-            return ret;
-
           ret = move_process_to_cgroup (pid, NULL, *path, err);
           if (UNLIKELY (ret < 0))
             return ret;
+
+          must_chown = true;
         }
 
       ret = enable_controllers (*path, err);
       if (UNLIKELY (ret < 0))
         return ret;
+
+      if (must_chown)
+        {
+          ret = chown_cgroups (*path, args->root_uid, args->root_gid, err);
+          if (UNLIKELY (ret < 0))
+            return ret;
+        }
     }
 
   if (cgroup_mode != CGROUP_MODE_UNIFIED && geteuid ())
@@ -1411,9 +1527,41 @@ libcrun_cgroup_enter_systemd (struct libcrun_cgroup_args *args, libcrun_error_t
 #endif
 }
 
+static inline void
+cleanup_sig_contp (void *p)
+{
+  pid_t *pp = p;
+  if (*pp < 0)
+    return;
+
+  TEMP_FAILURE_RETRY (kill (*pp, SIGCONT));
+}
+
+static bool
+must_stop_proc (runtime_spec_schema_config_linux_resources *resources)
+{
+  size_t i;
+
+  if (resources == NULL)
+    return false;
+
+  if (resources->cpu && (resources->cpu->cpus || resources->cpu->mems))
+    return true;
+
+  if (resources->unified)
+    {
+      for (i = 0; i < resources->unified->len; i++)
+        if (has_prefix (resources->unified->keys[i], "cpuset."))
+          return true;
+    }
+
+  return false;
+}
+
 int
 libcrun_cgroup_enter (struct libcrun_cgroup_args *args, libcrun_error_t *err)
 {
+  pid_t sigcont_cleanup __attribute__ ((cleanup (cleanup_sig_contp))) = -1;
   int cgroup_mode = args->cgroup_mode;
   char **path = args->path;
   int manager = args->manager;
@@ -1423,6 +1571,21 @@ libcrun_cgroup_enter (struct libcrun_cgroup_args *args, libcrun_error_t *err)
   int rootless;
   int ret;
 
+  /* If the cgroup configuration is limiting what CPUs/memory Nodes are available for the container,
+     then stop the container process during the cgroup configuration to avoid it being rescheduled on
+     a CPU that is not allowed.  This extra step is required for setting up the sub cgroup with the
+     systemd driver.  The alternative would be to temporarily setup the cpus/mems using d-bus.
+  */
+  if (must_stop_proc (args->resources))
+    {
+      ret = TEMP_FAILURE_RETRY (kill (args->pid, SIGSTOP));
+      if (UNLIKELY (ret < 0))
+        return crun_make_error (err, errno, "cannot stop container process '%d' with SIGSTOP", args->pid);
+
+      /* Send SIGCONT as soon as the function exits.  */
+      sigcont_cleanup = args->pid;
+    }
+
   if (cgroup_mode == CGROUP_MODE_HYBRID)
     {
       /* We don't really support hybrid mode, so check that cgroups2 is not using any controller.  */
@@ -1430,7 +1593,7 @@ libcrun_cgroup_enter (struct libcrun_cgroup_args *args, libcrun_error_t *err)
       size_t len;
       cleanup_free char *buffer = NULL;
 
-      ret = read_all_file ("/sys/fs/cgroup/unified/cgroup.controllers", &buffer, &len, err);
+      ret = read_all_file (CGROUP_ROOT "/unified/cgroup.controllers", &buffer, &len, err);
       if (UNLIKELY (ret < 0))
         return ret;
       if (len > 0)
@@ -1457,9 +1620,16 @@ libcrun_cgroup_enter (struct libcrun_cgroup_args *args, libcrun_error_t *err)
   if (LIKELY (ret >= 0))
     {
       if (cgroup_mode == CGROUP_MODE_UNIFIED && (root_uid != (uid_t) -1 || root_gid != (gid_t) -1))
-        return chown_cgroups (*path, root_uid, root_gid, err);
+        {
+          ret = chown_cgroups (*path, root_uid, root_gid, err);
+          if (UNLIKELY (ret < 0))
+              return ret;
+        }
 
-      return ret;
+      if (args->resources)
+        return libcrun_update_cgroup_resources (args->cgroup_mode, args->resources, *path, err);
+
+      return 0;
     }
 
   rootless = is_rootless (&tmp_err);
@@ -1496,12 +1666,18 @@ libcrun_cgroup_is_container_paused (const char *cgroup_path, int cgroup_mode, bo
   if (cgroup_mode == CGROUP_MODE_UNIFIED)
     {
       state = "1";
-      xasprintf (&path, "/sys/fs/cgroup/%s/cgroup.freeze", cgroup_path);
+
+      ret = append_paths (&path, err, CGROUP_ROOT, cgroup_path, "cgroup.freeze", NULL);
+      if (UNLIKELY (ret < 0))
+        return ret;
     }
   else
     {
       state = "FROZEN";
-      xasprintf (&path, "/sys/fs/cgroup/freezer/%s/freezer.state", cgroup_path);
+
+      ret = append_paths (&path, err, CGROUP_ROOT "/freezer", cgroup_path, "freezer.state", NULL);
+      if (UNLIKELY (ret < 0))
+        return ret;
     }
 
   ret = read_all_file (path, &content, NULL, err);
@@ -1520,15 +1696,22 @@ libcrun_cgroup_pause_unpause_with_mode (const char *cgroup_path, int cgroup_mode
   const char *state = "";
   int ret;
 
+  if (cgroup_path == NULL || cgroup_path[0] == '\0')
+    return crun_make_error (err, 0, "cannot %s the container without a cgroup", pause ? "pause" : "resume");
+
   if (cgroup_mode == CGROUP_MODE_UNIFIED)
     {
       state = pause ? "1" : "0";
-      xasprintf (&path, "/sys/fs/cgroup/%s/cgroup.freeze", cgroup_path);
+      ret = append_paths (&path, err, CGROUP_ROOT, cgroup_path, "cgroup.freeze", NULL);
+      if (UNLIKELY (ret < 0))
+        return ret;
     }
   else
     {
       state = pause ? "FROZEN" : "THAWED";
-      xasprintf (&path, "/sys/fs/cgroup/freezer/%s/freezer.state", cgroup_path);
+      ret = append_paths (&path, err, CGROUP_ROOT "/freezer", cgroup_path, "freezer.state", NULL);
+      if (UNLIKELY (ret < 0))
+        return ret;
     }
 
   ret = write_file (path, state, strlen (state), err);
@@ -1626,6 +1809,7 @@ libcrun_cgroup_read_pids (const char *path, bool recurse, pid_t **pids, libcrun_
   size_t n_pids, allocated;
   int dirfd;
   int mode;
+  int ret;
 
   if (path == NULL || *path == '\0')
     return 0;
@@ -1637,12 +1821,16 @@ libcrun_cgroup_read_pids (const char *path, bool recurse, pid_t **pids, libcrun_
   switch (mode)
     {
     case CGROUP_MODE_UNIFIED:
-      xasprintf (&cgroup_path, "/sys/fs/cgroup/%s", path);
+      ret = append_paths (&cgroup_path, err, CGROUP_ROOT, path, NULL);
+      if (UNLIKELY (ret < 0))
+        return ret;
       break;
 
     case CGROUP_MODE_HYBRID:
     case CGROUP_MODE_LEGACY:
-      xasprintf (&cgroup_path, "/sys/fs/cgroup/memory/%s", path);
+      ret = append_paths (&cgroup_path, err, CGROUP_ROOT "/memory", path, NULL);
+      if (UNLIKELY (ret < 0))
+        return ret;
       break;
 
     default:
@@ -1801,7 +1989,9 @@ libcrun_cgroup_destroy (const char *id, const char *path, const char *scope, int
         {
           cleanup_free char *cgroup_path = NULL;
 
-          xasprintf (&cgroup_path, "/sys/fs/cgroup/%s", path);
+          ret = append_paths (&cgroup_path, err, CGROUP_ROOT, path, NULL);
+          if (UNLIKELY (ret < 0))
+            return ret;
           ret = rmdir (cgroup_path);
           if (ret < 0 && errno == EBUSY)
             {
@@ -1819,7 +2009,9 @@ libcrun_cgroup_destroy (const char *id, const char *path, const char *scope, int
               if (mode == CGROUP_MODE_LEGACY && strcmp (subsystems[i], "unified") == 0)
                 continue;
 
-              xasprintf (&cgroup_path, "/sys/fs/cgroup/%s/%s", subsystems[i], path);
+              ret = append_paths (&cgroup_path, err, CGROUP_ROOT, subsystems[i], path, NULL);
+              if (UNLIKELY (ret < 0))
+                return ret;
 
               ret = rmdir (cgroup_path);
               if (ret < 0 && errno == EBUSY)
@@ -1896,7 +2088,7 @@ write_blkio_v2_resources_throttling (int fd, const char *name, throttling_s **th
     {
       int ret;
       size_t len;
-      len = sprintf (fmt_buf, "%" PRIu64 ":%" PRIu64 " %s=%lu\n", throttling[i]->major, throttling[i]->minor, name,
+      len = sprintf (fmt_buf, "%" PRIu64 ":%" PRIu64 " %s=%" PRIu64 "\n", throttling[i]->major, throttling[i]->minor, name,
                      throttling[i]->rate);
 
       ret = TEMP_FAILURE_RETRY (write (fd, fmt_buf, len));
@@ -1976,13 +2168,13 @@ write_blkio_resources (int dirfd, bool cgroup2, runtime_spec_schema_config_linux
 
           for (i = 0; i < blkio->weight_device_len; i++)
             {
-              len = sprintf (fmt_buf, "%lu:%lu %i\n", blkio->weight_device[i]->major, blkio->weight_device[i]->minor,
+              len = sprintf (fmt_buf, "%" PRIu64 ":%" PRIu64" %" PRIu16 "\n", blkio->weight_device[i]->major, blkio->weight_device[i]->minor,
                              blkio->weight_device[i]->weight);
               ret = TEMP_FAILURE_RETRY (write (w_device_fd, fmt_buf, len));
               if (UNLIKELY (ret < 0))
                 return crun_make_error (err, errno, "write blkio.weight_device");
 
-              len = sprintf (fmt_buf, "%lu:%lu %i\n", blkio->weight_device[i]->major, blkio->weight_device[i]->minor,
+              len = sprintf (fmt_buf, "%" PRIu64 ":%" PRIu64 " %" PRIu16 "\n", blkio->weight_device[i]->major, blkio->weight_device[i]->minor,
                              blkio->weight_device[i]->leaf_weight);
               ret = TEMP_FAILURE_RETRY (write (w_leafdevice_fd, fmt_buf, len));
               if (UNLIKELY (ret < 0))
@@ -1993,10 +2185,14 @@ write_blkio_resources (int dirfd, bool cgroup2, runtime_spec_schema_config_linux
   if (cgroup2)
     {
       cleanup_close int wfd = -1;
+      const char *name = "io.max";
 
-      wfd = openat (dirfd, "io.max", O_WRONLY);
+      wfd = openat (dirfd, name, O_WRONLY);
       if (UNLIKELY (wfd < 0))
-        return crun_make_error (err, errno, "open io.max");
+        {
+          ret = crun_make_error (err, errno, "open `%s`", name);
+          return check_cgroup_v2_controller_available_wrapper (ret, dirfd, name, err);
+        }
 
       ret = write_blkio_v2_resources_throttling (wfd, "rbps", (throttling_s **) blkio->throttle_read_bps_device,
                                                  blkio->throttle_read_bps_device_len, err);
@@ -2099,8 +2295,8 @@ write_hugetlb_resources (int dirfd, bool cgroup2,
 
       xasprintf (&filename, "hugetlb.%s.%s", htlb[i]->page_size, suffix);
 
-      len = sprintf (fmt_buf, "%lu", htlb[i]->limit);
-      ret = write_file_at (dirfd, filename, fmt_buf, len, err);
+      len = sprintf (fmt_buf, "%" PRIu64, htlb[i]->limit);
+      ret = write_file_and_check_controllers_at (cgroup2, dirfd, filename, fmt_buf, len, err);
       if (UNLIKELY (ret < 0))
         return ret;
     }
@@ -2134,14 +2330,14 @@ write_devices_resources_v1 (int dirfd, runtime_spec_schema_defs_linux_device_cgr
           char fmt_buf_major[16];
           char fmt_buf_minor[16];
 
-#define FMT_DEV(x, b)          \
-  do                           \
-    {                          \
-      if (x##_present)         \
-        sprintf (b, "%lu", x); \
-      else                     \
-        strcpy (b, "*");       \
-    }                          \
+#define FMT_DEV(x, b)               \
+  do                                \
+    {                               \
+      if (x##_present)              \
+        sprintf (b, "%" PRIi64, x); \
+      else                          \
+        strcpy (b, "*");            \
+    }                               \
   while (0)
 
           FMT_DEV (devs[i]->major, fmt_buf_major);
@@ -2289,7 +2485,7 @@ static int
 cg_itoa (char *buf, int64_t value, bool cgroup2)
 {
   if (! (cgroup2 && value == -1))
-    return sprintf (buf, "%lu", value);
+    return sprintf (buf, "%" PRIi64, value);
 
   memcpy (buf, "max", 4);
   return 3;
@@ -2333,8 +2529,7 @@ write_memory_swap (int dirfd, bool cgroup2, runtime_spec_schema_config_linux_res
 
   swap_buf_len = cg_itoa (swap_buf, swap, cgroup2);
 
-  return write_file_at (dirfd, cgroup2 ? "memory.swap.max" : "memory.memsw.limit_in_bytes", swap_buf, swap_buf_len,
-                        err);
+  return write_file_and_check_controllers_at (cgroup2, dirfd, cgroup2 ? "memory.swap.max" : "memory.memsw.limit_in_bytes", swap_buf, swap_buf_len, err);
 }
 
 static int
@@ -2382,7 +2577,7 @@ write_memory_resources (int dirfd, bool cgroup2, runtime_spec_schema_config_linu
       if (cgroup2)
         return crun_make_error (err, 0, "cannot set kernel memory with cgroupv2");
 
-      len = sprintf (fmt_buf, "%lu", memory->kernel);
+      len = sprintf (fmt_buf, "%" PRIu64, memory->kernel);
       ret = write_file_at (dirfd, "memory.kmem.limit_in_bytes", fmt_buf, len, err);
       if (UNLIKELY (ret < 0))
         return ret;
@@ -2390,8 +2585,8 @@ write_memory_resources (int dirfd, bool cgroup2, runtime_spec_schema_config_linu
 
   if (memory->reservation_present)
     {
-      len = sprintf (fmt_buf, "%lu", memory->reservation);
-      ret = write_file_at (dirfd, cgroup2 ? "memory.low" : "memory.soft_limit_in_bytes", fmt_buf, len, err);
+      len = sprintf (fmt_buf, "%" PRIu64, memory->reservation);
+      ret = write_file_and_check_controllers_at (cgroup2, dirfd, cgroup2 ? "memory.low" : "memory.soft_limit_in_bytes", fmt_buf, len, err);
       if (UNLIKELY (ret < 0))
         return ret;
     }
@@ -2409,7 +2604,7 @@ write_memory_resources (int dirfd, bool cgroup2, runtime_spec_schema_config_linu
       if (cgroup2)
         return crun_make_error (err, 0, "cannot set kernel TCP with cgroupv2");
 
-      len = sprintf (fmt_buf, "%lu", memory->kernel_tcp);
+      len = sprintf (fmt_buf, "%" PRIu64, memory->kernel_tcp);
       ret = write_file_at (dirfd, "memory.kmem.tcp.limit_in_bytes", fmt_buf, len, err);
       if (UNLIKELY (ret < 0))
         return ret;
@@ -2419,7 +2614,7 @@ write_memory_resources (int dirfd, bool cgroup2, runtime_spec_schema_config_linu
       if (cgroup2)
         return crun_make_error (err, 0, "cannot set memory swappiness with cgroupv2");
 
-      len = sprintf (fmt_buf, "%lu", memory->swappiness);
+      len = sprintf (fmt_buf, "%" PRIu64, memory->swappiness);
       ret = write_file_at (dirfd, "memory.swappiness", fmt_buf, len, err);
       if (UNLIKELY (ret < 0))
         return ret;
@@ -2439,7 +2634,7 @@ write_pids_resources (int dirfd, bool cgroup2, runtime_spec_schema_config_linux_
       int ret;
 
       len = cg_itoa (fmt_buf, pids->limit, cgroup2);
-      ret = write_file_at (dirfd, "pids.max", fmt_buf, len, err);
+      ret = write_file_and_check_controllers_at (cgroup2, dirfd, "pids.max", fmt_buf, len, err);
       if (UNLIKELY (ret < 0))
         return ret;
     }
@@ -2469,7 +2664,7 @@ write_cpu_resources (int dirfd_cpu, bool cgroup2, runtime_spec_schema_config_lin
 
       len = sprintf (fmt_buf, "%u", val);
 
-      ret = write_file_at (dirfd_cpu, cgroup2 ? "cpu.weight" : "cpu.shares", fmt_buf, len, err);
+      ret = write_file_and_check_controllers_at (cgroup2, dirfd_cpu, cgroup2 ? "cpu.weight" : "cpu.shares", fmt_buf, len, err);
       if (UNLIKELY (ret < 0))
         return ret;
     }
@@ -2479,7 +2674,7 @@ write_cpu_resources (int dirfd_cpu, bool cgroup2, runtime_spec_schema_config_lin
         period = cpu->period;
       else
         {
-          len = sprintf (fmt_buf, "%lu", cpu->period);
+          len = sprintf (fmt_buf, "%" PRIu64, cpu->period);
           ret = write_file_at (dirfd_cpu, "cpu.cfs_period_us", fmt_buf, len, err);
           if (UNLIKELY (ret < 0))
             return ret;
@@ -2491,7 +2686,7 @@ write_cpu_resources (int dirfd_cpu, bool cgroup2, runtime_spec_schema_config_lin
         quota = cpu->quota;
       else
         {
-          len = sprintf (fmt_buf, "%lu", cpu->quota);
+          len = sprintf (fmt_buf, "%" PRIu64, cpu->quota);
           ret = write_file_at (dirfd_cpu, "cpu.cfs_quota_us", fmt_buf, len, err);
           if (UNLIKELY (ret < 0))
             return ret;
@@ -2501,7 +2696,7 @@ write_cpu_resources (int dirfd_cpu, bool cgroup2, runtime_spec_schema_config_lin
     {
       if (cgroup2)
         return crun_make_error (err, 0, "realtime period not supported on cgroupv2");
-      len = sprintf (fmt_buf, "%lu", cpu->realtime_period);
+      len = sprintf (fmt_buf, "%" PRIu64, cpu->realtime_period);
       ret = write_file_at (dirfd_cpu, "cpu.rt_period_us", fmt_buf, len, err);
       if (UNLIKELY (ret < 0))
         return ret;
@@ -2510,7 +2705,7 @@ write_cpu_resources (int dirfd_cpu, bool cgroup2, runtime_spec_schema_config_lin
     {
       if (cgroup2)
         return crun_make_error (err, 0, "realtime runtime not supported on cgroupv2");
-      len = sprintf (fmt_buf, "%lu", cpu->realtime_runtime);
+      len = sprintf (fmt_buf, "%" PRIu64, cpu->realtime_runtime);
       ret = write_file_at (dirfd_cpu, "cpu.rt_runtime_us", fmt_buf, len, err);
       if (UNLIKELY (ret < 0))
         return ret;
@@ -2524,7 +2719,7 @@ write_cpu_resources (int dirfd_cpu, bool cgroup2, runtime_spec_schema_config_lin
         len = sprintf (fmt_buf, "max %" PRIi64, period);
       else
         len = sprintf (fmt_buf, "%" PRIi64 " %" PRIi64, quota, period);
-      ret = write_file_at (dirfd_cpu, "cpu.max", fmt_buf, len, err);
+      ret = write_file_and_check_controllers_at (cgroup2, dirfd_cpu, "cpu.max", fmt_buf, len, err);
       if (UNLIKELY (ret < 0))
         return ret;
     }
@@ -2532,20 +2727,20 @@ write_cpu_resources (int dirfd_cpu, bool cgroup2, runtime_spec_schema_config_lin
 }
 
 static int
-write_cpuset_resources (int dirfd_cpuset, int cgroup2 arg_unused, runtime_spec_schema_config_linux_resources_cpu *cpu,
+write_cpuset_resources (int dirfd_cpuset, int cgroup2, runtime_spec_schema_config_linux_resources_cpu *cpu,
                         libcrun_error_t *err)
 {
   int ret;
 
   if (cpu->cpus)
     {
-      ret = write_file_at (dirfd_cpuset, "cpuset.cpus", cpu->cpus, strlen (cpu->cpus), err);
+      ret = write_file_and_check_controllers_at (cgroup2, dirfd_cpuset, "cpuset.cpus", cpu->cpus, strlen (cpu->cpus), err);
       if (UNLIKELY (ret < 0))
         return ret;
     }
   if (cpu->mems)
     {
-      ret = write_file_at (dirfd_cpuset, "cpuset.mems", cpu->mems, strlen (cpu->mems), err);
+      ret = write_file_and_check_controllers_at (cgroup2, dirfd_cpuset, "cpuset.mems", cpu->mems, strlen (cpu->mems), err);
       if (UNLIKELY (ret < 0))
         return ret;
     }
@@ -2563,7 +2758,10 @@ update_cgroup_v1_resources (runtime_spec_schema_config_linux_resources *resource
       cleanup_close int dirfd_blkio = -1;
       runtime_spec_schema_config_linux_resources_block_io *blkio = resources->block_io;
 
-      xasprintf (&path_to_blkio, "/sys/fs/cgroup/blkio%s/", path);
+      ret = append_paths (&path_to_blkio, err, CGROUP_ROOT "/blkio", path, NULL);
+      if (UNLIKELY (ret < 0))
+        return ret;
+
       dirfd_blkio = open (path_to_blkio, O_DIRECTORY | O_RDONLY);
       if (UNLIKELY (dirfd_blkio < 0))
         return crun_make_error (err, errno, "open %s", path_to_blkio);
@@ -2581,8 +2779,13 @@ update_cgroup_v1_resources (runtime_spec_schema_config_linux_resources *resource
       cleanup_close int dirfd_netprio = -1;
       runtime_spec_schema_config_linux_resources_network *network = resources->network;
 
-      xasprintf (&path_to_netclass, "/sys/fs/cgroup/net_cls%s/", path);
-      xasprintf (&path_to_netprio, "/sys/fs/cgroup/net_prio%s/", path);
+      ret = append_paths (&path_to_netclass, err, CGROUP_ROOT "/net_cls", path, NULL);
+      if (UNLIKELY (ret < 0))
+        return ret;
+
+      ret = append_paths (&path_to_netprio, err, CGROUP_ROOT "/net_prio", path, NULL);
+      if (UNLIKELY (ret < 0))
+        return ret;
 
       dirfd_netclass = open (path_to_netclass, O_DIRECTORY | O_RDONLY);
       if (UNLIKELY (dirfd_netclass < 0))
@@ -2602,7 +2805,9 @@ update_cgroup_v1_resources (runtime_spec_schema_config_linux_resources *resource
       cleanup_free char *path_to_htlb = NULL;
       cleanup_close int dirfd_htlb = -1;
 
-      xasprintf (&path_to_htlb, "/sys/fs/cgroup/hugetlb%s/", path);
+      ret = append_paths (&path_to_htlb, err, CGROUP_ROOT "/hugetlb", path, NULL);
+      if (UNLIKELY (ret < 0))
+        return ret;
       dirfd_htlb = open (path_to_htlb, O_DIRECTORY | O_RDONLY);
       if (UNLIKELY (dirfd_htlb < 0))
         return crun_make_error (err, errno, "open %s", path_to_htlb);
@@ -2618,7 +2823,10 @@ update_cgroup_v1_resources (runtime_spec_schema_config_linux_resources *resource
       cleanup_free char *path_to_devs = NULL;
       cleanup_close int dirfd_devs = -1;
 
-      xasprintf (&path_to_devs, "/sys/fs/cgroup/devices%s/", path);
+      ret = append_paths (&path_to_devs, err, CGROUP_ROOT "/devices", path, NULL);
+      if (UNLIKELY (ret < 0))
+        return ret;
+
       dirfd_devs = open (path_to_devs, O_DIRECTORY | O_RDONLY);
       if (UNLIKELY (dirfd_devs < 0))
         return crun_make_error (err, errno, "open %s", path_to_devs);
@@ -2633,7 +2841,10 @@ update_cgroup_v1_resources (runtime_spec_schema_config_linux_resources *resource
       cleanup_free char *path_to_mem = NULL;
       cleanup_close int dirfd_mem = -1;
 
-      xasprintf (&path_to_mem, "/sys/fs/cgroup/memory%s/", path);
+      ret = append_paths (&path_to_mem, err, CGROUP_ROOT "/memory", path, NULL);
+      if (UNLIKELY (ret < 0))
+        return ret;
+
       dirfd_mem = open (path_to_mem, O_DIRECTORY | O_RDONLY);
       if (UNLIKELY (dirfd_mem < 0))
         return crun_make_error (err, errno, "open %s", path_to_mem);
@@ -2648,7 +2859,10 @@ update_cgroup_v1_resources (runtime_spec_schema_config_linux_resources *resource
       cleanup_free char *path_to_pid = NULL;
       cleanup_close int dirfd_pid = -1;
 
-      xasprintf (&path_to_pid, "/sys/fs/cgroup/pids%s/", path);
+      ret = append_paths (&path_to_pid, err, CGROUP_ROOT "/pids", path, NULL);
+      if (UNLIKELY (ret < 0))
+        return ret;
+
       dirfd_pid = open (path_to_pid, O_DIRECTORY | O_RDONLY);
       if (UNLIKELY (dirfd_pid < 0))
         return crun_make_error (err, errno, "open %s", path_to_pid);
@@ -2665,7 +2879,10 @@ update_cgroup_v1_resources (runtime_spec_schema_config_linux_resources *resource
       cleanup_free char *path_to_cpuset = NULL;
       cleanup_close int dirfd_cpuset = -1;
 
-      xasprintf (&path_to_cpu, "/sys/fs/cgroup/cpu%s/", path);
+      ret = append_paths (&path_to_cpu, err, CGROUP_ROOT "/cpu", path, NULL);
+      if (UNLIKELY (ret < 0))
+        return ret;
+
       dirfd_cpu = open (path_to_cpu, O_DIRECTORY | O_RDONLY);
       if (UNLIKELY (dirfd_cpu < 0))
         return crun_make_error (err, errno, "open %s", path_to_cpu);
@@ -2676,10 +2893,14 @@ update_cgroup_v1_resources (runtime_spec_schema_config_linux_resources *resource
       if (resources->cpu->cpus == NULL && resources->cpu->mems == NULL)
         return 0;
 
-      xasprintf (&path_to_cpuset, "/sys/fs/cgroup/cpuset%s/", path);
+      ret = append_paths (&path_to_cpuset, err, CGROUP_ROOT "/cpuset", path, NULL);
+      if (UNLIKELY (ret < 0))
+        return ret;
+
       dirfd_cpuset = open (path_to_cpuset, O_DIRECTORY | O_RDONLY);
       if (UNLIKELY (dirfd_cpuset < 0))
         return crun_make_error (err, errno, "open %s", path_to_cpuset);
+
       ret = write_cpuset_resources (dirfd_cpuset, false, resources->cpu, err);
       if (UNLIKELY (ret < 0))
         return ret;
@@ -2705,60 +2926,9 @@ write_unified_resources (int cgroup_dirfd, runtime_spec_schema_config_linux_reso
         return crun_make_error (err, 0, "key `%s` must be a file name without any slash", resources->unified->keys[i]);
 
       len = strlen (resources->unified->values[i]);
-      ret = write_file_at (cgroup_dirfd, resources->unified->keys[i], resources->unified->values[i], len, err);
+      ret = write_file_and_check_controllers_at (true, cgroup_dirfd, resources->unified->keys[i], resources->unified->values[i], len, err);
       if (UNLIKELY (ret < 0))
-        {
-          errno = crun_error_get_errno (err);
-
-          /* If the file is not found, try to give a more meaningful error message.  */
-          if (errno == ENOENT || errno == EPERM || errno == EACCES)
-            {
-              cleanup_free char *controllers = NULL;
-              libcrun_error_t tmp_err = NULL;
-              cleanup_free char *key = NULL;
-              char *saveptr = NULL;
-              bool found = false;
-              const char *token;
-              char *it;
-
-              /* Check if the specified controller is enabled.  */
-              key = xstrdup (resources->unified->keys[i]);
-
-              it = strchr (key, '.');
-              if (it == NULL)
-                {
-                  crun_error_release (err);
-                  return crun_make_error (err, 0, "the specified key has not the form CONTROLLER.VALUE `%s`",
-                                          resources->unified->keys[i]);
-                }
-              *it = '\0';
-
-              /* cgroup. files are not part of a controller.  Return the original error.  */
-              if (strcmp (key, "cgroup") == 0)
-                return ret;
-
-              /* If the cgroup.controllers file cannot be read, return the original error.  */
-              if (read_all_file_at (cgroup_dirfd, "cgroup.controllers", &controllers, NULL, &tmp_err) < 0)
-                {
-                  crun_error_release (&tmp_err);
-                  return ret;
-                }
-              for (token = strtok_r (controllers, " \n", &saveptr); token; token = strtok_r (NULL, " \n", &saveptr))
-                {
-                  if (strcmp (token, key) == 0)
-                    {
-                      found = true;
-                      break;
-                    }
-                }
-              if (! found)
-                {
-                  crun_error_release (err);
-                  return crun_make_error (err, 0, "the requested controller `%s` is not available", key);
-                }
-            }
-          return ret;
-        }
+        return ret;
     }
 
   return 0;
@@ -2774,7 +2944,9 @@ update_cgroup_v2_resources (runtime_spec_schema_config_linux_resources *resource
   if (resources->network)
     return crun_make_error (err, 0, "network limits not supported on cgroupv2");
 
-  xasprintf (&cgroup_path, "/sys/fs/cgroup%s", path);
+  ret = append_paths (&cgroup_path, err, CGROUP_ROOT, path, NULL);
+  if (UNLIKELY (ret < 0))
+    return ret;
 
   cgroup_dirfd = open (cgroup_path, O_DIRECTORY);
   if (UNLIKELY (cgroup_dirfd < 0))
