@@ -47,6 +47,10 @@
 #  include <dlfcn.h>
 #endif
 
+#ifdef HAVE_LIBKRUN
+#  include <libkrun.h>
+#endif
+
 #ifdef HAVE_SYSTEMD
 #  include <systemd/sd-daemon.h>
 #endif
@@ -79,6 +83,11 @@ struct container_entrypoint_s
 
   int hooks_out_fd;
   int hooks_err_fd;
+
+  /* If specified, it is called instead of
+     execve.  */
+  int (*exec_func) (void *container, void *arg, const char *pathname, char *const argv[]);
+  void *exec_func_arg;
 };
 
 struct sync_socket_message_s
@@ -394,6 +403,40 @@ sync_socket_send_sync (int fd, bool flush_errors, libcrun_error_t *err)
   return 0;
 }
 
+/*
+  Create an intermediate user namespace if there is a single id mapped
+  inside of the container user namespace and the container wants to run
+  with a different UID/GID than root.
+*/
+static bool
+need_intermediate_userns (runtime_spec_schema_config_schema *def)
+{
+  runtime_spec_schema_config_schema_process *process = def->process;
+  uid_t container_uid;
+  gid_t container_gid;
+
+  if (process == NULL)
+    return false;
+
+  container_uid = process->user ? process->user->uid : 0;
+  container_gid = process->user ? process->user->gid : 0;
+
+  if (container_uid == 0 && container_gid == 0)
+    return false;
+
+  if (def->linux->uid_mappings_len != 1 || def->linux->gid_mappings_len != 1)
+    return false;
+
+  if (def->linux->uid_mappings[0]->size != 1 || def->linux->gid_mappings[0]->size != 1)
+    return false;
+
+  if (def->linux->uid_mappings[0]->container_id == container_uid
+      && def->linux->gid_mappings[0]->container_id == container_gid)
+    return false;
+
+  return true;
+}
+
 static libcrun_container_t *
 make_container (runtime_spec_schema_config_schema *container_def)
 {
@@ -403,6 +446,8 @@ make_container (runtime_spec_schema_config_schema *container_def)
   container->host_uid = geteuid ();
   container->host_gid = getegid ();
 
+  container->use_intermediate_userns = need_intermediate_userns (container_def);
+
   return container;
 }
 
@@ -503,8 +548,7 @@ do_hooks (runtime_spec_schema_config_schema *def, pid_t pid, const char *id, boo
 {
   size_t i, stdin_len;
   int ret;
-  cleanup_free char *stdin = NULL;
-  const unsigned char *annotations = (const unsigned char *) "{}";
+  char *stdin = NULL;
   cleanup_free char *cwd_allocated = NULL;
   const char *rootfs = def->root ? def->root->path : "";
   yajl_gen gen = NULL;
@@ -516,14 +560,32 @@ do_hooks (runtime_spec_schema_config_schema *def, pid_t pid, const char *id, boo
         OOM ();
     }
 
+  gen = yajl_gen_alloc (NULL);
+  if (gen == NULL)
+    return crun_make_error (err, 0, "yajl_gen_alloc failed");
+
+  yajl_gen_map_open (gen);
+  yajl_gen_string (gen, YAJL_STR ("ociVersion"), strlen ("ociVersion"));
+  yajl_gen_string (gen, YAJL_STR ("1.0"), strlen ("1.0"));
+
+  yajl_gen_string (gen, YAJL_STR ("id"), strlen ("id"));
+  yajl_gen_string (gen, YAJL_STR (id), strlen (id));
+
+  yajl_gen_string (gen, YAJL_STR ("pid"), strlen ("pid"));
+  yajl_gen_integer (gen, pid);
+
+  yajl_gen_string (gen, YAJL_STR ("root"), strlen ("root"));
+  yajl_gen_string (gen, YAJL_STR (rootfs), strlen (rootfs));
+
+  yajl_gen_string (gen, YAJL_STR ("bundle"), strlen ("bundle"));
+  yajl_gen_string (gen, YAJL_STR (cwd), strlen (cwd));
+
+  yajl_gen_string (gen, YAJL_STR ("status"), strlen ("status"));
+  yajl_gen_string (gen, YAJL_STR (status), strlen (status));
+
   if (def && def->annotations && def->annotations->len)
     {
-      size_t len;
-
-      gen = yajl_gen_alloc (NULL);
-      if (gen == NULL)
-        return crun_make_error (err, 0, "yajl_gen_alloc failed");
-
+      yajl_gen_string (gen, YAJL_STR ("annotations"), strlen ("annotations"));
       yajl_gen_map_open (gen);
       for (i = 0; i < def->annotations->len; i++)
         {
@@ -534,14 +596,11 @@ do_hooks (runtime_spec_schema_config_schema *def, pid_t pid, const char *id, boo
           yajl_gen_string (gen, YAJL_STR (val), strlen (val));
         }
       yajl_gen_map_close (gen);
-
-      yajl_gen_get_buf (gen, &annotations, &len);
     }
 
-  stdin_len = xasprintf (&stdin,
-                         "{\"ociVersion\":\"1.0\", \"id\":\"%s\", \"pid\":%i, \"root\":\"%s\", \"bundle\":\"%s\", "
-                         "\"status\":\"%s\", \"annotations\":%s}",
-                         id, pid, rootfs, cwd, status, annotations);
+  yajl_gen_map_close (gen);
+
+  yajl_gen_get_buf (gen, (const unsigned char **) &stdin, &stdin_len);
 
   ret = 0;
 
@@ -567,6 +626,105 @@ do_hooks (runtime_spec_schema_config_schema *def, pid_t pid, const char *id, boo
   return ret;
 }
 
+#if HAVE_DLOPEN && HAVE_LIBKRUN
+static int
+libkrun_do_exec (void *container, void *arg, const char *pathname, char *const argv[])
+{
+  runtime_spec_schema_config_schema *def = ((libcrun_container_t *) container)->container_def;
+  int32_t (*krun_create_ctx) ();
+  int (*krun_start_enter) (uint32_t ctx_id);
+  int32_t (*krun_set_vm_config) (uint32_t ctx_id, uint8_t num_vcpus, uint32_t ram_mib);
+  int32_t (*krun_set_root) (uint32_t ctx_id, const char *root_path);
+  int32_t (*krun_set_exec) (uint32_t ctx_id, const char *exec_path, char *const argv[], char *const envp[]);
+  void *handle = arg;
+  uint32_t num_vcpus, ram_mib;
+  int32_t ctx_id, ret;
+  cpu_set_t set;
+
+  krun_create_ctx = dlsym (handle, "krun_create_ctx");
+  krun_start_enter = dlsym (handle, "krun_start_enter");
+  krun_set_vm_config = dlsym (handle, "krun_set_vm_config");
+  krun_set_root = dlsym (handle, "krun_set_root");
+  krun_set_exec = dlsym (handle, "krun_set_exec");
+  if (krun_create_ctx == NULL || krun_start_enter == NULL || krun_set_vm_config == NULL || krun_set_root == NULL
+      || krun_set_exec == NULL)
+    {
+      fprintf (stderr, "could not find symbol in `libkrun.so`");
+      dlclose (handle);
+      return -1;
+    }
+
+  /* If sched_getaffinity fails, default to 1 vcpu.  */
+  num_vcpus = 1;
+  /* If no memory limit is specified, default to 2G.  */
+  ram_mib = 2 * 1024;
+
+  if (def && def->linux && def->linux->resources && def->linux->resources->memory
+      && def->linux->resources->memory->limit_present)
+    ram_mib = def->linux->resources->memory->limit / (1024 * 1024);
+
+  CPU_ZERO (&set);
+  if (sched_getaffinity (getpid (), sizeof (set), &set) == 0)
+    num_vcpus = CPU_COUNT (&set);
+
+  ctx_id = krun_create_ctx ();
+  if (UNLIKELY (ctx_id < 0))
+    error (EXIT_FAILURE, -ret, "could not create krun context");
+
+  ret = krun_set_vm_config (ctx_id, num_vcpus, ram_mib);
+  if (UNLIKELY (ret < 0))
+    error (EXIT_FAILURE, -ret, "could not set krun vm configuration");
+
+  ret = krun_set_root (ctx_id, "/");
+  if (UNLIKELY (ret < 0))
+    error (EXIT_FAILURE, -ret, "could not set krun root");
+
+  ret = krun_set_exec (ctx_id, pathname, &argv[1], NULL);
+  if (UNLIKELY (ret < 0))
+    error (EXIT_FAILURE, -ret, "could not set krun executable");
+
+  return krun_start_enter (ctx_id);
+}
+#endif
+
+static int
+libcrun_configure_libkrun (struct container_entrypoint_s *args, libcrun_error_t *err)
+{
+#if HAVE_DLOPEN && HAVE_LIBKRUN
+  void *handle;
+#endif
+
+#if HAVE_DLOPEN && HAVE_LIBKRUN
+  handle = dlopen ("libkrun.so", RTLD_NOW);
+  if (handle == NULL)
+    return crun_make_error (err, 0, "could not load `libkrun.so`: %s", dlerror ());
+
+  args->exec_func = libkrun_do_exec;
+  args->exec_func_arg = handle;
+
+  return 0;
+#else
+  (void) args;
+  return crun_make_error (err, ENOTSUP, "libkrun or dlopen not present");
+#endif
+}
+
+static int
+libcrun_configure_handler (struct container_entrypoint_s *args, libcrun_error_t *err)
+{
+  const char *annotation;
+
+  annotation = find_annotation (args->container, "run.oci.handler");
+  /*Nothing to do.  */
+  if (annotation == NULL)
+    return 0;
+
+  if (strcmp (annotation, "krun") == 0)
+    return libcrun_configure_libkrun (args, err);
+
+  return crun_make_error (err, EINVAL, "invalid handler specified `%s`", annotation);
+}
+
 /* Initialize the environment where the container process runs.
    It is used by the container init process.  */
 static int
@@ -583,6 +741,10 @@ container_init_setup (void *args, char *notify_socket, int sync_socket, const ch
   cleanup_free char *rootfs = NULL;
   int no_new_privs;
 
+  ret = libcrun_configure_handler (args, err);
+  if (UNLIKELY (ret < 0))
+    return ret;
+
   ret = initialize_security (def->process, err);
   if (UNLIKELY (ret < 0))
     return ret;
@@ -591,11 +753,14 @@ container_init_setup (void *args, char *notify_socket, int sync_socket, const ch
   if (UNLIKELY (ret < 0))
     return ret;
 
-  rootfs = realpath (def->root->path, NULL);
-  if (UNLIKELY (rootfs == NULL))
+  if (def->root && def->root->path)
     {
-      /* If realpath failed for any reason, try the relative directory.  */
-      rootfs = xstrdup (def->root->path);
+      rootfs = realpath (def->root->path, NULL);
+      if (UNLIKELY (rootfs == NULL))
+        {
+          /* If realpath failed for any reason, try the relative directory.  */
+          rootfs = xstrdup (def->root->path);
+        }
     }
 
   if (entrypoint_args->terminal_socketpair[0] >= 0)
@@ -651,13 +816,16 @@ container_init_setup (void *args, char *notify_socket, int sync_socket, const ch
         return ret;
     }
 
-  ret = close_fds_ge_than (entrypoint_args->context->preserve_fds + 3, err);
+  ret = mark_for_close_fds_ge_than (entrypoint_args->context->preserve_fds + 3, err);
   if (UNLIKELY (ret < 0))
     crun_error_write_warning_and_release (entrypoint_args->context->output_handler_arg, &err);
 
-  ret = libcrun_do_pivot_root (container, entrypoint_args->context->no_pivot, rootfs, err);
-  if (UNLIKELY (ret < 0))
-    return ret;
+  if (rootfs)
+    {
+      ret = libcrun_do_pivot_root (container, entrypoint_args->context->no_pivot, rootfs, err);
+      if (UNLIKELY (ret < 0))
+        return ret;
+    }
 
   ret = libcrun_reopen_dev_null (err);
   if (UNLIKELY (ret < 0))
@@ -766,6 +934,13 @@ container_init_setup (void *args, char *notify_socket, int sync_socket, const ch
       close_and_reset (&entrypoint_args->seccomp_receiver_fd);
     }
 
+  if (entrypoint_args->container->use_intermediate_userns)
+    {
+      ret = libcrun_create_final_userns (entrypoint_args->container, err);
+      if (UNLIKELY (ret < 0))
+        return ret;
+    }
+
   capabilities = def->process ? def->process->capabilities : NULL;
   no_new_privs = def->process ? def->process->no_new_privileges : 1;
   ret = libcrun_set_caps (capabilities, container->container_uid, container->container_gid, no_new_privs, err);
@@ -913,6 +1088,13 @@ container_init (void *args, char *notify_socket, int sync_socket, libcrun_error_
       (void) lseek (2, 0, SEEK_END);
     }
 
+  if (entrypoint_args->exec_func)
+    {
+      ret = entrypoint_args->exec_func (entrypoint_args->container, entrypoint_args->exec_func_arg, exec_path,
+                                        def->process->args);
+      _exit (ret);
+    }
+
   execv (exec_path, def->process->args);
 
   if (errno == ENOENT)
@@ -925,8 +1107,9 @@ static int
 read_container_config_from_state (libcrun_container_t **container, const char *state_root, const char *id,
                                   libcrun_error_t *err)
 {
-  cleanup_free char *dir = NULL;
   cleanup_free char *config_file = NULL;
+  cleanup_free char *dir = NULL;
+  int ret;
 
   *container = NULL;
 
@@ -934,7 +1117,10 @@ read_container_config_from_state (libcrun_container_t **container, const char *s
   if (UNLIKELY (dir == NULL))
     return crun_make_error (err, 0, "cannot get state directory from `%s`", state_root);
 
-  xasprintf (&config_file, "%s/config.json", dir);
+  ret = append_paths (&config_file, err, dir, "config.json", NULL);
+  if (UNLIKELY (ret < 0))
+    return ret;
+
   *container = libcrun_container_load_from_file (config_file, err);
   if (*container == NULL)
     return crun_make_error (err, 0, "error loading `%s`", config_file);
@@ -1135,10 +1321,11 @@ write_container_status (libcrun_container_t *container, libcrun_context_t *conte
 {
   cleanup_free char *cwd = get_current_dir_name ();
   char *external_descriptors = libcrun_get_external_descriptors (container);
+  char *rootfs = container->container_def->root ? container->container_def->root->path : "";
   libcrun_container_status_t status = { .pid = pid,
                                         .cgroup_path = cgroup_path,
                                         .scope = scope,
-                                        .rootfs = container->container_def->root->path,
+                                        .rootfs = rootfs,
                                         .bundle = cwd,
                                         .created = created,
                                         .systemd_cgroup = context->systemd_cgroup,
@@ -1230,7 +1417,7 @@ wait_for_process (pid_t pid, libcrun_context_t *context, int terminal_fd, int no
     {
       char buf[12];
       size_t buf_len = sprintf (buf, "%d", pid);
-      ret = write_file (context->pid_file, buf, buf_len, err);
+      ret = write_file_with_flags (context->pid_file, O_CREAT | O_TRUNC, buf, buf_len, err);
       if (UNLIKELY (ret < 0))
         return ret;
     }
@@ -1273,13 +1460,20 @@ wait_for_process (pid_t pid, libcrun_context_t *context, int terminal_fd, int no
       state_root = libcrun_get_state_directory (context->state_root, context->id);
       if (UNLIKELY (state_root == NULL))
         return crun_make_error (err, 0, "cannot get state directory");
-      xasprintf (&oci_config_path, "%s/config.json", state_root);
+
+      ret = append_paths (&oci_config_path, err, state_root, "config.json", NULL);
+      if (UNLIKELY (ret < 0))
+        return ret;
 
       conf.runtime_root_path = state_root;
       conf.name = context->id;
       conf.bundle_path = context->bundle;
       conf.oci_config_path = oci_config_path;
 
+      ret = set_blocking_fd (seccomp_notify_fd, 0, err);
+      if (UNLIKELY (ret < 0))
+        return ret;
+
       ret = libcrun_load_seccomp_notify_plugins (&seccomp_notify_ctx, seccomp_notify_plugins, &conf, err);
       if (UNLIKELY (ret < 0))
         return ret;
@@ -1439,7 +1633,9 @@ open_seccomp_output (const char *id, int *fd, bool readonly, const char *state_r
   if (UNLIKELY (dir == NULL))
     return crun_make_error (err, 0, "cannot get state directory");
 
-  xasprintf (&dest_path, "%s/seccomp.bpf", dir);
+  ret = append_paths (&dest_path, err, dir, "seccomp.bpf", NULL);
+  if (UNLIKELY (ret < 0))
+    return ret;
 
   *fd = -1;
   if (readonly)
@@ -1598,6 +1794,8 @@ libcrun_container_run_internal (libcrun_container_t *container, libcrun_context_
     .hooks_out_fd = -1,
     .hooks_err_fd = -1,
     .seccomp_receiver_fd = -1,
+    .exec_func = context->exec_func,
+    .exec_func_arg = context->exec_func_arg,
   };
 
   if (def->hooks
@@ -1714,13 +1912,6 @@ libcrun_container_run_internal (libcrun_container_t *container, libcrun_context_
     ret = libcrun_cgroup_enter (&cg, err);
     if (UNLIKELY (ret < 0))
       return cleanup_watch (context, pid, sync_socket, terminal_fd, err);
-
-    if (def->linux && def->linux->resources)
-      {
-        ret = libcrun_update_cgroup_resources (cgroup_mode, def->linux->resources, cgroup_path, err);
-        if (UNLIKELY (ret < 0))
-          return cleanup_watch (context, pid, sync_socket, terminal_fd, err);
-      }
   }
 
   /* sync 1.  */
@@ -1832,14 +2023,17 @@ libcrun_container_run_internal (libcrun_container_t *container, libcrun_context_
 }
 
 static int
-check_config_file (runtime_spec_schema_config_schema *def, libcrun_error_t *err)
+check_config_file (runtime_spec_schema_config_schema *def, libcrun_context_t *context, libcrun_error_t *err)
 {
-  if (UNLIKELY (def->root == NULL))
-    return crun_make_error (err, 0, "invalid config file, no 'root' block specified");
   if (UNLIKELY (def->linux == NULL))
     return crun_make_error (err, 0, "invalid config file, no 'linux' block specified");
-  if (UNLIKELY (def->mounts == NULL))
-    return crun_make_error (err, 0, "invalid config file, no 'mounts' block specified");
+  if (context->exec_func == NULL)
+    {
+      if (UNLIKELY (def->root == NULL))
+        return crun_make_error (err, 0, "invalid config file, no 'root' block specified");
+      if (UNLIKELY (def->mounts == NULL))
+        return crun_make_error (err, 0, "invalid config file, no 'mounts' block specified");
+    }
   return 0;
 }
 
@@ -1856,7 +2050,9 @@ libcrun_copy_config_file (const char *id, const char *state_root, const char *co
   if (UNLIKELY (dir == NULL))
     return crun_make_error (err, 0, "cannot get state directory");
 
-  xasprintf (&dest_path, "%s/config.json", dir);
+  ret = append_paths (&dest_path, err, dir, "config.json", NULL);
+  if (UNLIKELY (ret < 0))
+    return ret;
 
   ret = read_all_file (config_file, &buffer, &len, err);
   if (UNLIKELY (ret < 0))
@@ -1890,7 +2086,7 @@ libcrun_container_run (libcrun_context_t *context, libcrun_container_t *containe
 
   container->context = context;
 
-  ret = check_config_file (def, err);
+  ret = check_config_file (def, context, err);
   if (UNLIKELY (ret < 0))
     return ret;
 
@@ -1992,7 +2188,7 @@ libcrun_container_create (libcrun_context_t *context, libcrun_container_t *conta
   if (def->oci_version && strstr (def->oci_version, "1.0") == NULL)
     return crun_make_error (err, 0, "unknown version specified");
 
-  ret = check_config_file (def, err);
+  ret = check_config_file (def, context, err);
   if (UNLIKELY (ret < 0))
     return ret;
 
@@ -2201,8 +2397,15 @@ libcrun_get_container_state_string (const char *id, libcrun_container_status_t *
       ret = libcrun_cgroup_is_container_paused (status->cgroup_path, cgroup_mode, &paused, err);
       if (UNLIKELY (ret < 0))
         {
-          /* The cgroup might have been cleaned up by systemd by the time we try to read it, so ignore ENOENT.  */
-          if (status->systemd_cgroup && crun_error_get_errno (err) == ENOENT)
+          /*
+            The cgroup might have been cleaned up by the time we try to read it, ignore both
+            ENOENT and ENODEV:
+            - ENOENT: if the open(CGROUP_PATH) fails because the cgroup was deleted.
+            - ENODEV: if the cgroup is deleted between the open and reading the freeze status.
+          */
+
+          errno = crun_error_get_errno (err);
+          if (errno == ENOENT || errno == ENODEV)
             {
               crun_error_release (err);
               *container_status = "stopped";
@@ -2291,7 +2494,10 @@ libcrun_container_state (libcrun_context_t *context, const char *id, FILE *out,
         goto exit;
       }
 
-    xasprintf (&config_file, "%s/config.json", dir);
+    ret = append_paths (&config_file, err, dir, "config.json", NULL);
+    if (UNLIKELY (ret < 0))
+      return ret;
+
     container = libcrun_container_load_from_file (config_file, err);
     if (UNLIKELY (container == NULL))
       {
@@ -2332,7 +2538,7 @@ int
 libcrun_container_exec (libcrun_context_t *context, const char *id, runtime_spec_schema_config_schema_process *process,
                         libcrun_error_t *err)
 {
-  int ret;
+  int container_status, ret;
   pid_t pid;
   libcrun_container_status_t status = {};
   const char *state_root = context->state_root;
@@ -2359,17 +2565,21 @@ libcrun_container_exec (libcrun_context_t *context, const char *id, runtime_spec
   ret = libcrun_is_container_running (&status, err);
   if (UNLIKELY (ret < 0))
     return ret;
+  container_status = ret;
 
   dir = libcrun_get_state_directory (state_root, id);
   if (UNLIKELY (dir == NULL))
     return crun_make_error (err, 0, "cannot get state directory");
 
-  xasprintf (&config_file, "%s/config.json", dir);
+  ret = append_paths (&config_file, err, dir, "config.json", NULL);
+  if (UNLIKELY (ret < 0))
+    return ret;
+
   container = libcrun_container_load_from_file (config_file, err);
   if (container == NULL)
     return crun_make_error (err, 0, "error loading config.json");
 
-  if (ret == 0)
+  if (container_status == 0)
     return crun_make_error (err, 0, "the container `%s` is not running.", id);
 
   ret = block_signals (err);
@@ -2503,6 +2713,10 @@ libcrun_container_exec (libcrun_context_t *context, const char *id, runtime_spec
             return ret;
         }
 
+      ret = mark_for_close_fds_ge_than (context->preserve_fds + 3, err);
+      if (UNLIKELY (ret < 0))
+        libcrun_fail_with_error ((*err)->status, "%s", (*err)->msg);
+
       if (! process->no_new_privileges)
         {
           ret = libcrun_apply_seccomp (seccomp_fd, seccomp_receiver_fd, seccomp_flags, seccomp_flags_len, err);
@@ -2528,10 +2742,6 @@ libcrun_container_exec (libcrun_context_t *context, const char *id, runtime_spec
             libcrun_fail_with_error ((*err)->status, "%s", (*err)->msg);
         }
 
-      ret = close_fds_ge_than (context->preserve_fds + 3, err);
-      if (UNLIKELY (ret < 0))
-        libcrun_fail_with_error ((*err)->status, "%s", (*err)->msg);
-
       if (process->no_new_privileges)
         {
           ret = libcrun_apply_seccomp (seccomp_fd, seccomp_receiver_fd, seccomp_flags, seccomp_flags_len, err);
@@ -2595,10 +2805,10 @@ libcrun_container_exec (libcrun_context_t *context, const char *id, runtime_spec
         }
     }
 
-  TEMP_FAILURE_RETRY (read (pipefd0, &b, sizeof (b)));
+  ret = TEMP_FAILURE_RETRY (read (pipefd0, &b, sizeof (b)));
   TEMP_FAILURE_RETRY (close (pipefd0));
   pipefd0 = -1;
-  if (b != '0')
+  if (ret != 1 || b != '0')
     ret = -1;
   else
     {
@@ -2788,7 +2998,7 @@ libcrun_container_restore (libcrun_context_t *context, const char *id, libcrun_c
   if (def->oci_version && strstr (def->oci_version, "1.0") == NULL)
     return crun_make_error (err, 0, "unknown version specified");
 
-  ret = check_config_file (def, err);
+  ret = check_config_file (def, context, err);
   if (UNLIKELY (ret < 0))
     return ret;
 
@@ -2845,13 +3055,6 @@ libcrun_container_restore (libcrun_context_t *context, const char *id, libcrun_c
     ret = libcrun_cgroup_enter (&cg, err);
     if (UNLIKELY (ret < 0))
       return ret;
-
-    if (def->linux && def->linux->resources)
-      {
-        ret = libcrun_update_cgroup_resources (cgroup_mode, def->linux->resources, cgroup_path, err);
-        if (UNLIKELY (ret < 0))
-          return ret;
-      }
   }
 
   get_current_timestamp (created);
@@ -2864,7 +3067,7 @@ libcrun_container_restore (libcrun_context_t *context, const char *id, libcrun_c
     {
       char buf[12];
       size_t buf_len = sprintf (buf, "%d", status.pid);
-      ret = write_file (context->pid_file, buf, buf_len, err);
+      ret = write_file_with_flags (context->pid_file, O_CREAT | O_TRUNC, buf, buf_len, err);
       if (UNLIKELY (ret < 0))
         return ret;
     }
