@@ -45,6 +45,9 @@ struct libcrun_context_s
   bool no_new_keyring;
   bool force_no_cgroup;
   bool no_pivot;
+
+  int (*exec_func) (void *container, void *arg, const char *pathname, char *const argv[]);
+  void *exec_func_arg;
 };
 
 enum
@@ -63,6 +66,8 @@ struct libcrun_container_s
   uid_t container_uid;
   gid_t container_gid;
 
+  bool use_intermediate_userns;
+
   void *private_data;
   struct libcrun_context_s *context;
 };
@@ -82,6 +87,7 @@ struct libcrun_checkpoint_restore_s
   bool shell_job;
   bool ext_unix_sk;
   bool detach;
+  const char *console_socket;
 };
 typedef struct libcrun_checkpoint_restore_s libcrun_checkpoint_restore_t;
 
