@@ -29,24 +29,42 @@
 #include <sys/mount.h>
 #include <sys/syscall.h>
 #include <sys/prctl.h>
-#include <sys/capability.h>
 #include <sys/time.h>
 #include <sys/resource.h>
 #include <sys/sysmacros.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <seccomp.h>
+
+#ifdef HAVE_SECCOMP
+# include <seccomp.h>
+#endif
 #include <linux/seccomp.h>
 #include <linux/filter.h>
 #include <sys/prctl.h>
 #include <sys/syscall.h>
 
+#ifndef __NR_seccomp
+#  define __NR_seccomp 0xffff // seccomp syscall number unknown for this architecture
+#endif
+
+#ifndef SECCOMP_SET_MODE_STRICT
+#  define SECCOMP_SET_MODE_STRICT 0
+#endif
+
+#ifndef SECCOMP_SET_MODE_FILTER
+#  define SECCOMP_SET_MODE_FILTER 1
+#endif
+
+#ifndef SECCOMP_FILTER_FLAG_TSYNC
+#  define SECCOMP_FILTER_FLAG_TSYNC (1UL << 0)
+#endif
+
 #ifndef SECCOMP_FILTER_FLAG_LOG
-# define SECCOMP_FILTER_FLAG_LOG (1UL << 1)
+#  define SECCOMP_FILTER_FLAG_LOG (1UL << 1)
 #endif
 
 #ifndef SECCOMP_FILTER_FLAG_SPEC_ALLOW
-# define SECCOMP_FILTER_FLAG_SPEC_ALLOW	(1UL << 2)
+#  define SECCOMP_FILTER_FLAG_SPEC_ALLOW (1UL << 2)
 #endif
 
 static int
@@ -55,9 +73,10 @@ syscall_seccomp (unsigned int operation, unsigned int flags, void *args)
   return (int) syscall (__NR_seccomp, operation, flags, args);
 }
 
-unsigned long
+static unsigned long
 get_seccomp_operator (const char *name, libcrun_error_t *err)
 {
+#ifdef HAVE_SECCOMP
   if (strcmp (name, "SCMP_CMP_NE") == 0)
     return SCMP_CMP_NE;
   if (strcmp (name, "SCMP_CMP_LT") == 0)
@@ -75,24 +94,56 @@ get_seccomp_operator (const char *name, libcrun_error_t *err)
 
   crun_make_error (err, 0, "seccomp get operator", name);
   return 0;
+#else
+  return 0;
+#endif
 }
 
-unsigned long long
-get_seccomp_action (const char *name, libcrun_error_t *err)
+static unsigned long long
+get_seccomp_action (const char *name, int errno_ret, libcrun_error_t *err)
 {
-  if (strcmp (name, "SCMP_ACT_KILL") == 0)
-    return SCMP_ACT_KILL;
-  if (strcmp (name, "SCMP_ACT_ALLOW") == 0)
-    return SCMP_ACT_ALLOW;
-  if (strcmp (name, "SCMP_ACT_TRAP") == 0)
-    return SCMP_ACT_TRAP;
-  if (strcmp (name, "SCMP_ACT_ERRNO") == 0)
-    return SCMP_ACT_ERRNO (EPERM);
-  if (strcmp (name, "SCMP_ACT_TRACE") == 0)
-    return SCMP_ACT_TRACE (EPERM);
+#ifdef HAVE_SECCOMP
+  const char *p;
 
+  p = name;
+  if (strncmp (p, "SCMP_ACT_", 9))
+    goto fail;
+
+  p += 9;
+
+  if (strcmp (p, "ALLOW") == 0)
+    return SCMP_ACT_ALLOW;
+  else if (strcmp (p, "ERRNO") == 0)
+    return SCMP_ACT_ERRNO (errno_ret);
+  else if (strcmp (p, "KILL") == 0)
+    return SCMP_ACT_KILL;
+# ifdef SCMP_ACT_LOG
+  else if (strcmp (p, "LOG") == 0)
+    return SCMP_ACT_LOG;
+# endif
+  else if (strcmp (p, "TRAP") == 0)
+    return SCMP_ACT_TRAP;
+  else if (strcmp (p, "TRACE") == 0)
+    return SCMP_ACT_TRACE (errno_ret);
+# ifdef SCMP_ACT_KILL_PROCESS
+  else if (strcmp (p, "KILL_PROCESS") == 0)
+    return SCMP_ACT_KILL_PROCESS;
+# endif
+# ifdef SCMP_ACT_KILL_THREAD
+  else if (strcmp (p, "KILL_THREAD") == 0)
+    return SCMP_ACT_KILL_THREAD;
+# endif
+# ifdef SCMP_ACT_NOTIFY
+  else if (strcmp (p, "NOTIFY") == 0)
+    return SCMP_ACT_NOTIFY;
+# endif
+
+fail:
   crun_make_error (err, 0, "seccomp get action", name);
   return 0;
+#else
+  return 0;
+#endif
 }
 
 static void
@@ -108,15 +159,19 @@ make_lowercase (char *str)
 static void
 cleanup_seccompp (void *p)
 {
+#ifdef HAVE_SECCOMP
   scmp_filter_ctx *ctx = (void **) p;
   if (*ctx)
     seccomp_release (*ctx);
+#endif
 }
-#define cleanup_seccomp __attribute__((cleanup (cleanup_seccompp)))
+#define cleanup_seccomp __attribute__ ((cleanup (cleanup_seccompp)))
 
 int
-libcrun_apply_seccomp (int infd, char **seccomp_flags, size_t seccomp_flags_len, libcrun_error_t *err)
+libcrun_apply_seccomp (int infd, int listener_receiver_fd, char **seccomp_flags, size_t seccomp_flags_len,
+                       libcrun_error_t *err)
 {
+#ifdef HAVE_SECCOMP
   int ret;
   struct sock_fprog seccomp_filter;
   cleanup_free char *bpf = NULL;
@@ -127,19 +182,18 @@ libcrun_apply_seccomp (int infd, char **seccomp_flags, size_t seccomp_flags_len,
     return 0;
 
   if (UNLIKELY (lseek (infd, 0, SEEK_SET) == (off_t) -1))
-    return crun_make_error (err, 0, "lseek");
-
+    return crun_make_error (err, errno, "lseek");
 
   /* if no seccomp flag was specified use a sane default.  */
   if (seccomp_flags == NULL)
-    flags = SECCOMP_FILTER_FLAG_LOG|SECCOMP_FILTER_FLAG_SPEC_ALLOW;
+    flags = SECCOMP_FILTER_FLAG_LOG | SECCOMP_FILTER_FLAG_SPEC_ALLOW;
   else
     {
       size_t i = 0;
       for (i = 0; i < seccomp_flags_len; i++)
         {
           if (strcmp (seccomp_flags[i], "SECCOMP_FILTER_FLAG_TSYNC") == 0)
-              flags |= SECCOMP_FILTER_FLAG_TSYNC;
+            flags |= SECCOMP_FILTER_FLAG_TSYNC;
           else if (strcmp (seccomp_flags[i], "SECCOMP_FILTER_FLAG_SPEC_ALLOW") == 0)
             flags |= SECCOMP_FILTER_FLAG_SPEC_ALLOW;
           else if (strcmp (seccomp_flags[i], "SECCOMP_FILTER_FLAG_LOG") == 0)
@@ -156,6 +210,15 @@ libcrun_apply_seccomp (int infd, char **seccomp_flags, size_t seccomp_flags_len,
   seccomp_filter.len = len / 8;
   seccomp_filter.filter = (struct sock_filter *) bpf;
 
+  if (listener_receiver_fd >= 0)
+    {
+# ifdef SECCOMP_FILTER_FLAG_NEW_LISTENER
+      flags |= SECCOMP_FILTER_FLAG_NEW_LISTENER;
+# else
+      return crun_make_error (err, 0, "SECCOMP_FILTER_FLAG_NEW_LISTENER not supported");
+# endif
+    }
+
   ret = syscall_seccomp (SECCOMP_SET_MODE_FILTER, flags, &seccomp_filter);
   if (UNLIKELY (ret < 0))
     {
@@ -166,12 +229,24 @@ libcrun_apply_seccomp (int infd, char **seccomp_flags, size_t seccomp_flags_len,
         return crun_make_error (err, errno, "seccomp (SECCOMP_SET_MODE_FILTER)");
     }
 
+  if (listener_receiver_fd >= 0)
+    {
+      int fd = ret;
+
+      ret = send_fd_to_socket (listener_receiver_fd, fd, err);
+      if (UNLIKELY (ret < 0))
+        return crun_error_wrap (err, "send listener fd `%d` to receiver", fd);
+    }
   return 0;
+#else
+  return 0;
+#endif
 }
 
 int
 libcrun_generate_seccomp (libcrun_container_t *container, int outfd, unsigned int options, libcrun_error_t *err)
 {
+#ifdef HAVE_SECCOMP
   runtime_spec_schema_config_linux_seccomp *seccomp = container->container_def->linux->seccomp;
   int ret;
   size_t i;
@@ -189,8 +264,8 @@ libcrun_generate_seccomp (libcrun_container_t *container, int outfd, unsigned in
   if (seccomp->default_action != NULL)
     def_action = seccomp->default_action;
 
-  default_action = get_seccomp_action (def_action, err);
-  if (UNLIKELY (default_action == 0))
+  default_action = get_seccomp_action (def_action, EPERM, err);
+  if (UNLIKELY (err && *err != NULL))
     return crun_make_error (err, 0, "invalid seccomp action `%s`", seccomp->default_action);
 
   ctx = seccomp_init (default_action);
@@ -207,23 +282,28 @@ libcrun_generate_seccomp (libcrun_container_t *container, int outfd, unsigned in
         arch += 10;
       stpncpy (lowercase_arch, arch, sizeof (lowercase_arch));
       make_lowercase (lowercase_arch);
-#ifdef SECCOMP_ARCH_RESOLVE_NAME
+# ifdef SECCOMP_ARCH_RESOLVE_NAME
       arch_token = seccomp_arch_resolve_name (lowercase_arch);
       if (arch_token == 0)
         return crun_make_error (err, 0, "seccomp unknown architecture %s", arch);
-#else
+# else
       arch_token = SCMP_ARCH_NATIVE;
-#endif
+# endif
       ret = seccomp_arch_add (ctx, arch_token);
       if (ret < 0 && ret != -EEXIST)
-        return crun_make_error (err, 0, "seccomp adding architecture");
+        return crun_make_error (err, -ret, "seccomp adding architecture");
     }
 
   for (i = 0; i < seccomp->syscalls_len; i++)
     {
       size_t j;
-      action = get_seccomp_action (seccomp->syscalls[i]->action, err);
-      if (UNLIKELY (action == 0))
+      int errno_ret = EPERM;
+
+      if (seccomp->syscalls[i]->errno_ret_present)
+        errno_ret = seccomp->syscalls[i]->errno_ret;
+
+      action = get_seccomp_action (seccomp->syscalls[i]->action, errno_ret, err);
+      if (UNLIKELY (err && *err != NULL))
         return crun_make_error (err, 0, "invalid seccomp action `%s`", seccomp->syscalls[i]->action);
 
       if (action == default_action)
@@ -253,9 +333,7 @@ libcrun_generate_seccomp (libcrun_container_t *container, int outfd, unsigned in
               size_t k;
               struct scmp_arg_cmp arg_cmp[6];
               bool multiple_args = false;
-              uint32_t count[6];
-
-              memset (count, 0, sizeof (count));
+              uint32_t count[6] = {};
 
               for (k = 0; k < seccomp->syscalls[i]->args_len && k < 6; k++)
                 {
@@ -287,13 +365,9 @@ libcrun_generate_seccomp (libcrun_container_t *container, int outfd, unsigned in
 
               if (! multiple_args)
                 {
-                  ret = seccomp_rule_add_array (ctx,
-                                                action,
-                                                syscall,
-                                                k,
-                                                arg_cmp);
+                  ret = seccomp_rule_add_array (ctx, action, syscall, k, arg_cmp);
                   if (UNLIKELY (ret < 0))
-                    return crun_make_error (err, 0, "seccomp_rule_add_array");
+                    return crun_make_error (err, -ret, "seccomp_rule_add_array");
                 }
               else
                 {
@@ -301,13 +375,9 @@ libcrun_generate_seccomp (libcrun_container_t *container, int outfd, unsigned in
 
                   for (r = 0; r < k; r++)
                     {
-                      ret = seccomp_rule_add_array (ctx,
-                                                    action,
-                                                    syscall,
-                                                    1,
-                                                    &arg_cmp[r]);
+                      ret = seccomp_rule_add_array (ctx, action, syscall, 1, &arg_cmp[r]);
                       if (UNLIKELY (ret < 0))
-                        return crun_make_error (err, 0, "seccomp_rule_add_array");
+                        return crun_make_error (err, -ret, "seccomp_rule_add_array");
                     }
                 }
             }
@@ -318,8 +388,11 @@ libcrun_generate_seccomp (libcrun_container_t *container, int outfd, unsigned in
     {
       ret = seccomp_export_bpf (ctx, outfd);
       if (UNLIKELY (ret < 0))
-        return crun_make_error (err, 0, "seccomp_export_bpf");
+        return crun_make_error (err, -ret, "seccomp_export_bpf");
     }
 
   return 0;
+#else
+  return 0;
+#endif
 }
