@@ -40,16 +40,29 @@
 #include <sys/vfs.h>
 #include <linux/magic.h>
 #include <limits.h>
+#include <stdarg.h>
 
+#ifndef CLOSE_RANGE_CLOEXEC
+# define CLOSE_RANGE_CLOEXEC (1U << 2)
+#endif
 #ifndef RESOLVE_IN_ROOT
 #  define RESOLVE_IN_ROOT 0x10
 #endif
+#ifndef __NR_close_range
+#  define __NR_close_range 436
+#endif
 #ifndef __NR_openat2
 #  define __NR_openat2 437
 #endif
 
 #define MAX_READLINKS 32
 
+static int
+syscall_close_range (unsigned int fd, unsigned int max_fd, unsigned int flags)
+{
+  return (int) syscall (__NR_close_range, fd, max_fd, flags);
+}
+
 static int
 syscall_openat2 (int dirfd, const char *path, uint64_t flags, uint64_t mode, uint64_t resolve)
 {
@@ -111,9 +124,9 @@ write_file_at (int dirfd, const char *name, const void *data, size_t len, libcru
 }
 
 int
-write_file (const char *name, const void *data, size_t len, libcrun_error_t *err)
+write_file_with_flags (const char *name, int flags, const void *data, size_t len, libcrun_error_t *err)
 {
-  cleanup_close int fd = open (name, O_WRONLY | O_CREAT, 0700);
+  cleanup_close int fd = open (name, O_WRONLY | flags, 0700);
   int ret;
   if (UNLIKELY (fd < 0))
     return crun_make_error (err, errno, "opening file `%s` for writing", name);
@@ -125,6 +138,12 @@ write_file (const char *name, const void *data, size_t len, libcrun_error_t *err
   return ret;
 }
 
+int
+write_file (const char *name, const void *data, size_t len, libcrun_error_t *err)
+{
+  return write_file_with_flags (name, O_CREAT, data, len, err);
+}
+
 int
 detach_process ()
 {
@@ -223,48 +242,39 @@ ensure_directory_internal_at (int dirfd, char *path, size_t len, int mode, libcr
 {
   char *it = path + len;
   int ret = 0;
-  int parent_created;
+  bool parent_created = false;
 
-  for (parent_created = 0; parent_created < 2; parent_created++)
+  for (;;)
     {
       ret = mkdirat (dirfd, path, mode);
-      if (ret == 0)
-        break;
+      if (ret == 0 || errno == EEXIST)
+        return 0;
 
-      if (errno == EEXIST)
-        {
-          ret = 0;
-          break;
-        }
-
-      if (errno != ENOENT || parent_created)
+      if (parent_created || errno != ENOENT)
         {
           /* On errors check if the directory already exists.  */
           ret = crun_dir_p (path, false, err);
           if (ret > 0)
-            break;
+            return 0;
 
           return crun_make_error (err, errno, "create directory `%s`", path);
         }
-      else
-        {
-          while (it > path && *it != '/')
-            {
-              it--;
-              len--;
-            }
-          if (it == path)
-            {
-              ret = 0;
-              break;
-            }
 
-          *it = '\0';
-          ret = ensure_directory_internal_at (dirfd, path, len - 1, mode, err);
-          *it = '/';
-          if (UNLIKELY (ret < 0))
-            break;
+      while (it > path && *it != '/')
+        {
+          it--;
+          len--;
         }
+      if (it == path)
+        return 0;
+
+      *it = '\0';
+      ret = ensure_directory_internal_at (dirfd, path, len - 1, mode, err);
+      *it = '/';
+      if (UNLIKELY (ret < 0))
+        return ret;
+
+      parent_created = true;
     }
   return ret;
 }
@@ -335,7 +345,7 @@ safe_openat (int dirfd, const char *rootfs, size_t rootfs_len, const char *path,
         {
           if (errno == ENOSYS)
             openat2_supported = false;
-          if (errno == ENOSYS || errno == EINVAL)
+          if (errno == ENOSYS || errno == EINVAL || errno == EPERM)
             goto fallback;
           return crun_make_error (err, errno, "openat2 `%s`", path);
         }
@@ -1358,6 +1368,10 @@ run_process_with_stdin_timeout_envp (char *path, char **args, const char *cwd, i
       char *tmp_args[] = { path, NULL };
       int dev_null_fd = -1;
 
+      ret = mark_for_close_fds_ge_than (3, err);
+      if (UNLIKELY (ret < 0))
+        libcrun_fail_with_error ((*err)->status, "%s", (*err)->msg);
+
       if (out_fd < 0 || err_fd < 0)
         {
           dev_null_fd = open ("/dev/null", O_WRONLY);
@@ -1392,7 +1406,7 @@ run_process_with_stdin_timeout_envp (char *path, char **args, const char *cwd, i
 }
 
 int
-close_fds_ge_than (int n, libcrun_error_t *err)
+mark_for_close_fds_ge_than (int n, libcrun_error_t *err)
 {
   cleanup_close int cfd = -1;
   cleanup_dir DIR *dir = NULL;
@@ -1401,6 +1415,12 @@ close_fds_ge_than (int n, libcrun_error_t *err)
   struct statfs sfs;
   struct dirent *next;
 
+  ret = syscall_close_range (n, UINT_MAX, CLOSE_RANGE_CLOEXEC);
+  if (ret == 0)
+    return 0;
+  if (ret < 0 && errno != EINVAL && errno != ENOSYS)
+    return crun_make_error (err, errno, "close_range from %d", n);
+
   cfd = open ("/proc/self/fd", O_DIRECTORY | O_RDONLY | O_CLOEXEC);
   if (UNLIKELY (cfd < 0))
     return crun_make_error (err, errno, "open /proc/self/fd");
@@ -1896,3 +1916,83 @@ safe_write (int fd, const void *buf, ssize_t count)
     }
   return written;
 }
+
+int
+append_paths (char **out, libcrun_error_t *err, ...)
+{
+  const size_t MAX_PARTS = 32;
+  const char *parts[MAX_PARTS];
+  size_t sizes[MAX_PARTS];
+  size_t total_len = 0;
+  size_t n_parts = 0;
+  size_t copied = 0;
+  va_list ap;
+  size_t i;
+
+  va_start (ap, err);
+  for (;;)
+    {
+      const char *part;
+      size_t size;
+
+      part = va_arg (ap, const char *);
+      if (part == NULL)
+        break;
+
+      if (n_parts == MAX_PARTS)
+        {
+          va_end (ap);
+          return crun_make_error (err, EINVAL, "too many paths specified");
+        }
+
+      if (n_parts == 0)
+        {
+          /* For the first component allow only one '/'.  */
+          while (part[0] == '/' && part[1] == '/')
+            part++;
+        }
+      else
+        {
+          /* And drop any initial '/' for other components.  */
+          while (part[0] == '/')
+            part++;
+        }
+
+      size = strlen (part);
+      if (size == 0)
+        continue;
+
+      while (size > 1 && part[size - 1] == '/')
+        size--;
+
+      parts[n_parts] = part;
+      sizes[n_parts] = size;
+
+      n_parts++;
+    }
+  va_end (ap);
+
+  total_len = n_parts + 1;
+  for (i = 0; i < n_parts; i++)
+    total_len += sizes[i];
+
+  *out = xmalloc (total_len);
+
+  copied = 0;
+  for (i = 0; i < n_parts; i++)
+    {
+      bool has_trailing_slash;
+
+      has_trailing_slash = copied > 0 && (*out)[copied - 1] == '/';
+      if (i > 0 && !has_trailing_slash)
+        {
+          (*out)[copied] = '/';
+          copied += 1;
+        }
+
+      memcpy (*out + copied, parts[i], sizes[i]);
+      copied += sizes[i];
+    }
+  (*out)[copied] = '\0';
+  return 0;
+}
