@@ -149,6 +149,8 @@ int xasprintf (char **str, const char *fmt, ...);
 
 int crun_path_exists (const char *path, libcrun_error_t *err);
 
+int write_file_with_flags (const char *name, int flags, const void *data, size_t len, libcrun_error_t *err);
+
 int write_file (const char *name, const void *data, size_t len, libcrun_error_t *err);
 
 int write_file_at (int dirfd, const char *name, const void *data, size_t len, libcrun_error_t *err);
@@ -212,7 +214,7 @@ size_t format_default_id_mapping (char **ret, uid_t container_id, uid_t host_id,
 int run_process_with_stdin_timeout_envp (char *path, char **args, const char *cwd, int timeout, char **envp,
                                          char *stdin, size_t stdin_len, int out_fd, int err_fd, libcrun_error_t *err);
 
-int close_fds_ge_than (int n, libcrun_error_t *err);
+int mark_for_close_fds_ge_than (int n, libcrun_error_t *err);
 
 void get_current_timestamp (char *out);
 
@@ -245,6 +247,8 @@ int safe_openat (int dirfd, const char *rootfs, size_t rootfs_len, const char *p
 
 ssize_t safe_write (int fd, const void *buf, ssize_t count);
 
+int append_paths (char **out, libcrun_error_t *err, ...);
+
 LIBCRUN_PUBLIC int libcrun_str2sig (const char *name);
 
 #endif
