@@ -17,8 +17,6 @@
  */
 #define _GNU_SOURCE
 
-#include <config.h>
-
 #include <unistd.h>
 #include <stdlib.h>
 #include <stdio.h>
@@ -138,11 +136,19 @@ int main (int argc, char **argv)
   if (strcmp (argv[1], "echo") == 0)
     {
       if (argc < 3)
-        error (EXIT_FAILURE, 0, "'cat' requires an argument");
+        error (EXIT_FAILURE, 0, "'echo' requires an argument");
       fputs (argv[2], stdout);
       exit (0);
     }
 
+  if (strcmp (argv[1], "printenv") == 0)
+    {
+      if (argc < 3)
+        error (EXIT_FAILURE, 0, "'printenv' requires an argument");
+      fputs (getenv (argv[2]), stdout);
+      exit (0);
+    }
+
   if (strcmp (argv[1], "cat") == 0)
     {
       if (argc < 3)
@@ -172,10 +178,9 @@ int main (int argc, char **argv)
 
   if (strcmp (argv[1], "gethostname") == 0)
     {
-      char buffer[64];
+      char buffer[64] = {};
       int ret;
 
-      memset (buffer, 0, sizeof (buffer));
       ret = gethostname (buffer, sizeof (buffer) - 1);
       if (ret < 0)
         error (EXIT_FAILURE, errno, "gethostname");
@@ -204,9 +209,14 @@ int main (int argc, char **argv)
     }
   if (strcmp (argv[1], "pause") == 0)
     {
+      unsigned int remaining = 120;
+
       close (1);
       close (2);
-      pause ();
+
+      while (remaining)
+        remaining = sleep (remaining);
+
       exit (0);
     }
   if (strcmp (argv[1], "forkbomb") == 0)
