@@ -1,4 +1,4 @@
-#!/bin/env $PYTHON
+#!/bin/env python3
 # crun - OCI runtime written in C
 #
 # Copyright (C) 2017, 2018, 2019 Giuseppe Scrivano <giuseppe@scrivano.org>
@@ -45,6 +45,28 @@ def helper_mount(options):
         return [m.vfs_options, m.fs_options]
     return -1
 
+def test_mount_symlink():
+    conf = base_config()
+    conf['process']['args'] = ['/init', 'cat', '/proc/self/mountinfo']
+    add_all_namespaces(conf)
+    mount_opt = {"destination": "/etc/localtime", "type": "bind", "source": "/etc/localtime", "options": ["bind", "ro"]}
+    conf['mounts'].append(mount_opt)
+    out, _ = run_and_get_output(conf, hide_stderr=True)
+    if "Rome" in out:
+        return 0
+    return -1
+
+def test_mount_symlink_not_existing():
+    conf = base_config()
+    conf['process']['args'] = ['/init', 'cat', '/proc/self/mountinfo']
+    add_all_namespaces(conf)
+    mount_opt = {"destination": "/etc/not-existing", "type": "bind", "source": "/etc/localtime", "options": ["bind", "ro"]}
+    conf['mounts'].append(mount_opt)
+    out, _ = run_and_get_output(conf, hide_stderr=True)
+    if "foo/bar" in out:
+        return 0
+    return -1
+
 def test_mount_ro():
     a = helper_mount("ro")[0]
     if "ro" in a:
@@ -116,6 +138,8 @@ all_tests = {
     "test-mount-nosuid" : test_mount_nosuid,
     "test-mount-sync" : test_mount_sync,
     "test-mount-dirsync" : test_mount_dirsync,
+    "test-mount-symlink" : test_mount_symlink,
+    "test-mount-symlink-not-existing" : test_mount_symlink_not_existing,
 }
 
 if __name__ == "__main__":
